from abc import ABC, abstractmethod
from enum import Enum
from multiprocessing import Process, Pipe
from multiprocessing.connection import Connection
from random import choice
from typing import Optional, Sequence

import msgpack  # type: ignore
import networkit as nk  # type: ignore
from hoa.ast.acceptance import AcceptanceAtom, AtomType  # type: ignore
from hoa.ast.boolean_expression import PositiveAnd, PositiveOr  # type: ignore
from hoa.core import Edge, State  # type: ignore
from networkit.graph import Graph  # type: ignore

from .drivers import Driver, UserDriver
from .hoa import Automaton, Transition, fmt_edge, fmt_state, parse


class Action(ABC):
    @abstractmethod
    def run(self, runner: "SingleRunner") -> None:
        raise NotImplementedError

    def __str__(self) -> str:
        return type(self).__name__


class Condition(ABC):
    @abstractmethod
    def check(self, runner: "SingleRunner"):
        pass

    def __str__(self) -> str:
        return type(self).__name__


class StopRunner(Exception):
    pass


class Runner(ABC):
    @abstractmethod
    def init(self):
        raise NotImplementedError

    @abstractmethod
    def step(self, values=Optional[set]):
        raise NotImplementedError

    @abstractmethod
    def add_transition_hook(self, hook):
        raise NotImplementedError

    @abstractmethod
    def add_nondet_action(self, action):
        raise NotImplementedError

    @abstractmethod
    def add_deadlock_action(self, action):
        raise NotImplementedError

    @property
    @abstractmethod
    def count(self):
        raise NotImplementedError


def spawn_runner(pipe: Connection):  # type: ignore
    fname, monitor = pipe.recv(), pipe.recv()
    aut = parse(fname)
    runner = SingleRunner(aut, UserDriver([]), monitor)
    while True:
        data = pipe.recv_bytes()
        msg = msgpack.loads(data)
        match msg:
            case "INIT":
                runner.init()
            case "QUIT":
                break
            case "GET_COUNT":
                pipe.send(runner.count)
            case set(values):
                x = runner.step(values)
                pipe.send_bytes(msgpack.dumps(x))
                # pipe.send(x)
            case ("ADD_HOOK", hook):
                runner.add_transition_hook(hook)
            case ("ADD_NONDET", action):
                x = runner.add_nondet_action(action)
            case ("ADD_DEADLOCK", action):
                x = runner.add_deadlock_action(action)


class MPCompositeRunner(Runner):
    INIT_MSG = msgpack.dumps("INIT")
    QUIT_MSG = msgpack.dumps("QUIT")
    GET_COUNT_MSG = msgpack.dumps("GET_COUNT")

    def __init__(self, automata: Sequence[Automaton], drv: Driver,
                 monitor: bool = False) -> None:
        self.driver = drv
        self.runners: list[Runner] = []
        self.procs = []
        for a in automata:
            parent_pipe, child_pipe = Pipe()
            proc = Process(target=spawn_runner, args=(child_pipe,))
            self.procs.append((proc, parent_pipe))
            parent_pipe.send(a.filename)
            parent_pipe.send(monitor)
            proc.start()

    def __del__(self):
        for proc, pipe in self.procs:
            pipe.close()
            proc.kill()

    @property
    def count(self):
        _, pipe = self.procs[0]
        pipe.send_bytes(MPCompositeRunner.GET_COUNT_MSG)
        return pipe.recv()

    def init(self):
        for _, pipe in self.procs:
            pipe.send_bytes(MPCompositeRunner.INIT_MSG)

    def step(self, inputs: Optional[set] = None):
        values = inputs or self.driver.get()
        msg = msgpack.dumps(values)
        for _, pipe in self.procs:
            pipe.send_bytes(msg)
        result = [tr for _, pipe in self.procs for tr in msgpack.loads(pipe.recv_bytes())]  # noqa: E501
        return result

    def add_transition_hook(self, hook):
        msg = msgpack.dumps(("ADD_HOOK", hook))
        for _, pipe in self.procs:
            pipe.send_bytes(msg)

    def add_nondet_action(self, action):
        msg = msgpack.dumps(("ADD_NONDET", action))
        for _, pipe in self.procs:
            pipe.send_bytes(msg)

    def add_deadlock_action(self, action):
        msg = msgpack.dumps(("ADD_DEADLOCK", action))
        for _, pipe in self.procs:
            pipe.send_bytes(msg)


class CompositeRunner(Runner):
    """Runner for multiple automata, fed by the same driver."""
    def __init__(self, automata: Sequence[Automaton], drv: Driver,
                 monitor: bool = False) -> None:
        self.driver = drv
        self.runners: list[Runner] = []
        for a in automata:
            if all(
                x in a.hoa.header.properties
                for x in ("complete", "deterministic")
            ):
                self.runners.append(DetCompleteSingleRunner(a, drv, monitor))
            else:
                self.runners.append(SingleRunner(a, drv, monitor))

    @property
    def count(self):
        return self.runners[0].count

    def init(self) -> None:
        for runner in self.runners:
            runner.init()

    def step(self, inputs: Optional[set] = None) -> list[Transition]:
        values = inputs or self.driver.get()
        result = [tr for r in self.runners for tr in r.step(values)]
        return result

    def add_transition_hook(self, hook):
        for runner in self.runners:
            runner.add_transition_hook(hook)

    def add_nondet_action(self, action):
        for runner in self.runners:
            runner.add_nondet_action(action)

    def add_deadlock_action(self, action):
        for runner in self.runners:
            runner.add_deadlock_action(action)


class SingleRunner(Runner):
    """Runner for a single automaton."""
    def __init__(self, aut: Automaton, drv: Driver, mon: bool = False) -> None:
        self.aut = aut
        self.aps = list(aut.get_aps())
        self.driver = drv
        self.state: int | None = None
        self.count = 0
        # self.trace = []
        self.deadlock_actions: list[Action] = []
        self.nondet_actions: list[Action] = []
        self.transition_hooks: list[Hook] = []
        self.candidates: list[Edge] = []
        prp = self.aut.hoa.header.properties or []
        self.deterministic = "deterministic" in prp
        # TODO make this configurable
        if mon:
            chk = AcceptanceChecker.make_checker(self.aut)
            self.transition_hooks.append(Hook(chk, Reset()))

    @property
    def count(self):
        return self._count

    @count.setter
    def count(self, value: int):
        self._count = value

    def init(self) -> None:
        # TODO support initial state conjunction (alternating automata)
        for x in self.aut.hoa.header.start_states:
            for y in x:
                self.state = y
                return

    def add_transition_hook(self, hook):
        self.transition_hooks.append(hook)

    def add_nondet_action(self, action):
        self.nondet_actions.append(action)

    def add_deadlock_action(self, action):
        self.deadlock_actions.append(action)

    def step(self, inputs: Optional[set] = None) -> list[Transition]:
        """return False iff automaton stuttered"""
        assert self.state is not None
        if inputs is None:
            inputs = self.driver.get()
        if self.deterministic:
            candidate = self.aut.get_first_candidate(self.state, inputs)
            self.candidates = [candidate] if candidate is not None else []
        else:
            self.candidates = list(self.aut.get_candidates(self.state, inputs))
        if not self.candidates:
            for action in self.deadlock_actions:
                action.run(self)
            return []
        elif len(self.candidates) > 1:
            for action in self.nondet_actions:
                action.run(self)
        if len(self.candidates) >= 1:
            edge = self.candidates[0]
            self.candidates = []
            old_state, next_state = self.state, edge.state_conj[0]

            self.count += 1
            self.state = next_state
            for hook in self.transition_hooks:
                hook.run(self)
            return [(old_state, inputs, next_state)]
        return []


class DetCompleteSingleRunner(SingleRunner):
    """Optimized `SingleRunner` for deterministic complete automata."""
    def step(self, inputs: Optional[set] = None) -> list[Transition]:
        assert self.state is not None
        if inputs is None:
            inputs = self.driver.get()
        edge = self.aut.get_first_candidate(self.state, inputs)
        next_state = edge.state_conj[0]
        self.count += 1
        old_state, self.state = self.state, next_state
        for hook in self.transition_hooks:
            hook.run(self)
        return [(old_state, inputs, next_state)]


class Reach(Condition):
    """Condition triggered by reaching a target state."""
    def __str__(self) -> str:
        return f"Reach {self.target}"

    def __init__(self, target: int) -> None:
        self.target = target

    def check(self, runner: SingleRunner):
        return runner.state == self.target


class Bound(Condition):
    """Condition triggered when an execution reaches a certain length."""
    def __str__(self) -> str:
        return f"Bound: {self.bound}"

    def __init__(self, bound) -> None:
        self.bound = bound

    def check(self, runner: SingleRunner):
        return runner.count > self.bound


class Always(Condition):
    """Trivial condition, always triggered."""
    def check(self, _):
        return True


class Hook:
    """A Hook is a combination of a triggering condition and an action."""
    def __init__(self, condition: Condition, action: Action) -> None:
        self.condition = condition
        self.action = action

    def run(self, runner: SingleRunner):
        if self.condition.check(runner):
            msg = f"Hook: {self.condition} triggered {self.action} at step {runner.count}"  # noqa: E501
            print(msg)
            self.action.run(runner)


class Reset(Action):
    """Reset the runner to its initial state."""
    def run(self, runner: SingleRunner) -> None:
        runner.init()


class Log(Action):
    """Write a message to standard output."""
    def __init__(self, msg: str) -> None:
        self.msg = msg

    def __str__(self) -> str:
        return f"Log {self.msg}"

    def run(self, runner: SingleRunner) -> None:
        print(f"{self.msg} at {fmt_state(runner.state)}")


class PressEnter(Action):
    """Ask for user confirmation before proceeding."""
    def run(self, _) -> None:
        input("Press [Enter] to continue...")


class RandomChoice(Action):
    """Randomly choose a candidate."""
    def run(self, runner: SingleRunner) -> None:
        chosen = choice(runner.candidates)
        runner.candidates = [chosen]


class UserChoice(Action):
    """Let the user choose a candidate."""
    def run(self, runner: SingleRunner) -> None:
        for i, edge in enumerate(runner.candidates):
            print(f"[{i}]\t{fmt_edge(edge, runner.aps)}")
        choice_int = -1
        while not 0 <= choice_int < len(runner.candidates):
            choice = input("Choose a transition from above: ")
            choice_int = int(choice) if choice.isdecimal() else -1
        chosen = runner.candidates[choice_int]
        runner.candidates = [chosen]


class Quit(Action):
    """Terminate the runner (and HOAX)."""
    def __init__(self, cause=None):
        super().__init__()
        self.cause = cause

    def run(self, runner):
        raise StopRunner()


class PrefixType(Enum):
    GOOD = 1
    BAD = 2
    UGLY = 3

    def invert(self) -> "PrefixType":
        if self == PrefixType.GOOD:
            return PrefixType.BAD
        if self == PrefixType.BAD:
            return PrefixType.GOOD
        return self


class AcceptanceChecker(Condition):
    """Checks for satisfaction of an acceptance condition.

    For details, see: https://doi.org/10.48550/arXiv.2507.11126
    """
    @abstractmethod
    def check(self, runner: SingleRunner) -> PrefixType | None:
        pass

    @staticmethod
    def make_checker(aut: Automaton) -> "AcceptanceChecker":
        """Generate an acceptance checker from the automaton's acceptance \
            condition.

        Args:
            aut (Automaton): A HOA automaton

        Returns:
            AcceptanceChecker: An acceptance checker
        """
        acond = aut.hoa.header.acceptance.condition
        all_states = aut.states

        def get_uglies(accept: set[int]):
            result = set()
            for k_id in aut.cond.iterNodes():
                if aut.cond.degree(k_id) > 0:
                    continue
                k = set(aut.graph_node2scc[k_id][1])
                k_minus_accept = k - accept
                if not k_minus_accept:
                    continue
                # Build graph for k_minus_accept
                graph = Graph(directed=True)

                for e in aut.graph.iterEdges():
                    if e[0] in k_minus_accept and e[1] in k_minus_accept:
                        graph.addEdge(*e, addMissing=True)
                for x in accept:
                    graph.removeNode(x)

                # Check whether k_minus_accept is acyclic
                seen = set()
                for node in k_minus_accept:
                    # Skip nodes visited in a previous DFS
                    if node in seen:
                        continue
                    dfs = []

                    def callback(u):
                        dfs.append(u)
                        for v in graph.iterNeighbors(u):
                            dfs.extend(v)
                            seen.add(v)

                    nk.graph.Traversal.DFSfrom(graph, node, callback)
                    if len(dfs) != len(set(dfs)):
                        result.add(k_minus_accept)
                        break
            return result

        def _mk(cond):
            def get_acceptance_set(index: int):
                accept = aut.acc_sets[index]
                if cond.negated:
                    accept = all_states - accept
                return frozenset(accept)
            match cond:
                case AcceptanceAtom(atom_type=AtomType.INFINITE):
                    accept = get_acceptance_set(cond.acceptance_set)
                    return Inf(accept, aut, get_uglies(accept))
                case AcceptanceAtom(atom_type=AtomType.FINITE):
                    accept = get_acceptance_set(cond.acceptance_set)
                    return Fin(accept, aut, get_uglies(accept))
                case PositiveAnd():
                    return And([_mk(c) for c in cond.operands])
                case PositiveOr():
                    return Or([_mk(c) for c in cond.operands])
        chk = _mk(acond)
        chk.set_filename(aut.filename)
        return chk


class BaseChecker(AcceptanceChecker):
    """Base class for Fin/Inf checkers."""
    def __init__(self, aset: set[int], aut: Automaton, uglies):
        self.aset = aset
        self.aut = aut
        self.uglies = uglies
        self.name: str | None = None
        self.cache: dict = {}

    def set_filename(self, name: str) -> None:
        self.name = name

    @abstractmethod
    def check_state(self, state: int) -> PrefixType | None:
        raise NotImplementedError

    def check(self, runner: SingleRunner) -> PrefixType | None:
        assert runner.state is not None
        try:
            return self.cache[runner.state]
        except KeyError:
            value = self.check_state(runner.state)
            self.cache[runner.state] = value
            return value


class Inf(BaseChecker):
    def __str__(self):
        return f"Inf({{{self.aset}}}){'@' if self.name else ""}{self.name or ""}"  # noqa: E501

    def check_state(self, state: int) -> PrefixType | None:
        t, t_minimal = self.aut.get_trap_set_of(state)
        if t <= self.aset:
            return PrefixType.GOOD
        if not (t & self.aset):
            return PrefixType.BAD
        if t_minimal:
            return (
                PrefixType.UGLY
                if any(t & u for u in self.uglies)
                else PrefixType.GOOD)
        return None


class Fin(BaseChecker):
    def __str__(self):
        return f"Fin({{{self.aset}}}){'@' if self.name else ""}{self.name or ""}"  # noqa: E501

    def check_state(self, state: int) -> PrefixType | None:
        t, t_minimal = self.aut.get_trap_set_of(state)
        if t <= self.aset:
            return PrefixType.BAD
        if not (t & self.aset):
            return PrefixType.GOOD
        if t_minimal:
            return (
                PrefixType.UGLY
                if any(t & u for u in self.uglies)
                else PrefixType.BAD)
        return None


class Neg(AcceptanceChecker):
    def __init__(self, mon: AcceptanceChecker):
        self.mon = mon

    def __str__(self):
        if isinstance(self.mon, Inf):
            return f"Fin({{{self.mon.aset}}})"
        return f"!({self.mon})"

    def check(self, runner: SingleRunner):
        p = self.mon.check(runner)
        if p is not None:
            return p.invert()
        return None


class And(AcceptanceChecker):
    def __init__(self, monitors):
        self.monitors = monitors

    def __str__(self):
        return " & ".join(f"({m})" for m in self.monitors)

    def check(self, runner: SingleRunner):
        checks = set()
        for m in self.monitors:
            check = m.check(runner)
            if check == PrefixType.UGLY:
                return PrefixType.UGLY
            checks.add(check)
        # If prefix is bad for at least one operand, it's bad
        if PrefixType.BAD in checks:
            return PrefixType.BAD
        if PrefixType.GOOD in checks:
            return PrefixType.GOOD
        return None


class Or(AcceptanceChecker):
    def __init__(self, monitors):
        self.monitors = monitors

    def __str__(self):
        return " | ".join(f"({m})" for m in self.monitors)

    def check(self, runner: SingleRunner):
        checks = set()
        for m in self.monitors:
            check = m.check(runner)
            checks.add(check)
        # If prefix is good for at least one operand, it's good
        if PrefixType.GOOD in checks:
            return PrefixType.GOOD
        # Prefix is bad for all operands = it's bad
        if PrefixType.BAD in checks and PrefixType.UGLY not in checks:
            return PrefixType.BAD
        # Prefix is bad for some operands and ugly for others = it's ugly
        if PrefixType.UGLY not in checks:
            return PrefixType.UGLY
        return None



import logging
from abc import ABC, abstractmethod
from io import TextIOWrapper
from json import loads
from random import choices

log = logging.getLogger(__name__)


class EndOfFiniteTrace(Exception):
    pass


class Driver(ABC):
    def __init__(self, aps: list[str]) -> None:
        log.debug(f"Instantiating {type(self).__name__} for {aps=}")
        self.aps = aps

    @abstractmethod
    def get(self) -> set:
        raise NotImplementedError


class CompositeDriver(Driver):
    def __init__(self) -> None:
        self.aps: list[str] = []
        self.drivers: list[Driver] = []

    def append(self, driver: Driver):
        self.drivers.append(driver)
        assert all(ap not in self.aps for ap in driver.aps), \
            "Some APs have multiple drivers: " \
            f"{[ap for ap in self.aps if ap in driver.aps]}"
        self.aps.extend(driver.aps)

    def get(self) -> set:
        result = set()
        for d in self.drivers:
            result |= d.get()
        return result


class UserDriver(Driver):
    def __init__(self, aps: list[str]) -> None:
        super().__init__(aps)

    @staticmethod
    def handle(in_str: str) -> bool:
        in_str = in_str.strip().lower()
        try:
            in_int = int(in_str)
            return bool(in_int)
        except ValueError:
            pass
        return False if in_str in ('false', 'null', '') else bool(in_str)

    def get(self):
        result = set()
        for ap in self.aps:
            value = handle(input(f"{ap}? "))
            if value:
                result.add(ap)
        return result


class RandomDriver(Driver):
    pop = True, False

    def __init__(self, aps) -> None:
        self.aps = aps
        self.k = len(self.aps)
        self.cum_weights: tuple[float, int] | None = None

    def get(self) -> set:
        result = choices(self.pop, k=self.k, cum_weights=self.cum_weights)
        return set(ap for ap, value in zip(self.aps, result) if value)


class StreamDriver(Driver):
    def __init__(self, aps, stream: TextIOWrapper, diff: bool = False) -> None:
        super().__init__(aps)
        self.diff = diff
        self.stream = stream

    def get(self) -> set:
        return self.read_next()

    def read_next(self) -> set:
        raise NotImplementedError


class JSONDriver(StreamDriver):
    def read_next(self) -> set:
        line = self.stream.readline()
        return set(x for x, y in loads(line) if y)


class SimpleTxtDriver(StreamDriver):
    def read_next(self):
        line = self.stream.readline()
        if not line:
            self.stream.close()
            raise EndOfFiniteTrace(self.stream.name)
        data = set()
        while line:
            if line == "\n":
                return data
            ap = line[:-1]
            if ap in self.aps:
                data.add(ap)
            line = self.stream.readline()
        return data


def handle(in_str: str) -> bool:
    in_str = in_str.strip().lower()
    try:
        in_int = int(in_str)
        return bool(in_int)
    except ValueError:
        pass
    return False if in_str in ('false', 'null', '') else bool(in_str)



from collections import defaultdict
from pathlib import Path
from sys import intern
from typing import Any, Collection, Iterable, Optional, Sequence

import hoa.ast.boolean_expression as ast  # type: ignore
import hoa.ast.label as ast_label  # type: ignore
import networkit as nk  # type: ignore
from networkit import Graph
from hoa.core import HOA, Edge, State  # type: ignore
from hoa.parsers import HOAParser  # type: ignore


get_first_candidate = intern("get_first_candidate")
get_trap_set_of = intern("get_trap_set_of")


def fmt_state(s: State) -> str:
    if s.name is not None:
        return f"{s.index} ('{s.name}')"
    return str(s.index)


def fmt_edge(e: Edge, aps: list[str]) -> str:
    lbl = fmt_expr(e.label, aps)
    tgt = ' '.join(str(i) for i in e.state_conj)
    return f"{lbl} --> {tgt}"


def extract_aps(all_aps: Sequence[str], aps: Collection[str | int]) -> list[str]:  # noqa: E501
    """Return only the propositions from `all_aps` that are in `aps` \
        (by name or index)."""
    tmp = (i if type(i) is str else all_aps[int(i)] for i in aps)
    return [i for i in tmp if i in all_aps]


def fmt_expr(node, aps: list[str]) -> str:
    """Format an expression into a string."""
    def recurse_and_reduce(op: str):
        return op.join(fmt_expr(x, aps) for x in node.operands)
    if isinstance(node, ast_label.LabelAtom):
        return aps[node.proposition]
    if isinstance(node, ast.FalseFormula):
        return "false"
    if isinstance(node, ast.TrueFormula):
        return "true"
    if isinstance(node, ast.And):
        return recurse_and_reduce(" & ")
    if isinstance(node, ast.Or):
        return recurse_and_reduce(" | ")
    if isinstance(node, ast.Not):
        return "!" + fmt_expr(node.argument, aps)
    raise Exception(f"Unexpected node {node}")


Transition = tuple[int, set, int]
"""A transition is a triple (source state, valuation, target state)"""


class Automaton:
    def __init__(self, aut: HOA, filename: Optional[str] = None, ctrl: Optional[set] = None):  # noqa: E501
        self.hoa = aut
        self.ctrl = ctrl or set()
        self.cache: dict[tuple, Any] = {}
        self.filename = filename
        self.states = max(x.index for x in aut.body.state2edges)
        self.int2edges: list[Sequence[Edge]] = [()] * (self.states + 1)
        for x, edges in aut.body.state2edges.items():
            self.int2edges[x.index] = edges

        self.graph, self.cond, self.graph_node2scc, self.cond_node2scc = self.graph_and_cond()  # noqa: E501

        # Compute accepting sets
        self.acc_sets = defaultdict(set)
        for x in self.hoa.body.state2edges:
            for s in (x.acc_sig or tuple()):
                self.acc_sets[s].add(x.index)

    def get_trap_set_of(self, index: int) -> tuple[set, bool]:
        key = (get_trap_set_of, index)
        try:
            return self.cache[key]
        except KeyError:
            k_id, k_nodes = self.graph_node2scc[index]
            is_minimal = self.cond.degree(k_id) == 0
            result = set(k_nodes)

            def callback(u):
                result.update(self.cond_node2scc[u])

            nk.graph.Traversal.DFSfrom(self.cond, k_id, callback)
            self.cache[key] = result, is_minimal
            return result, is_minimal

    def get_aps(self):
        """Yield all the atomic propositions of the automaton."""
        yield from self.hoa.header.propositions

    def get_states(self):
        pass

    def get_initial_states(self):
        yield from self.hoa.header.start_states

    def get_edges(self, index: int) -> Sequence[Edge]:
        """Yield all the edges from state `index`."""
        return self.int2edges[index]

    def evaluate(self, node, valuation: set) -> bool:
        """Evaluate an expression `node` against `valuation`."""
        match node:
            case ast_label.LabelAtom():
                return self.hoa.header.propositions[node.proposition] in valuation  # noqa: E501
            case ast.FalseFormula():
                return False
            case ast.TrueFormula():
                return True
            case ast.And(operands=ops):
                return all(self.evaluate(x, valuation) for x in ops)
            case ast.Or(operands=ops):
                return any(self.evaluate(x, valuation) for x in ops)
            case ast.Not(argument=arg):
                return not self.evaluate(arg, valuation)
        raise Exception(f"Unexpected node {node}")

    def get_candidates(self, index: int, values: set) -> Iterable[Edge]:
        """Yield all candidate edges rooted in `index` for a given valuation.

        Args:
            index (int): A state in the automaton.
            values (set): A valuation.

        Yields:
            Iterable[Edge]: Edges that `valuation` may trigger.
        """
        for edge in self.get_edges(index):
            if self.evaluate(edge.label, values):
                yield edge

    def get_first_candidate(self, index: int, values: set):
        """Return the first potential edge for the valuation `values`"""
        key = (get_first_candidate, index, *sorted(values))
        try:
            return self.cache[key]
        except KeyError:
            for edge in self.get_edges(index):
                if self.evaluate(edge.label, values):
                    self.cache[key] = edge
                    return edge
            self.cache[key] = None
            return self.cache[key]

    def graph_and_cond(self) -> tuple[Graph, Graph, dict, dict]:
        # Build digraph of automaton
        g = nk.Graph(directed=True)
        for i, edges in enumerate(self.int2edges):
            if edges is None:
                continue
            for e in edges:
                g.addEdge(i, e.state_conj[0], addMissing=True)
        # Compute SCCs
        sccs = nk.components.StronglyConnectedComponents(g)
        sccs.run()
        # Compute condensation of g
        cond = nk.Graph(directed=True)
        g_node2scc = {}
        cond_node2scc = {}
        for k in sccs.getComponents():
            s = cond.addNode()
            for node in k:
                g_node2scc[node] = s, k
                cond_node2scc[s] = k

        for (src, tgt) in g.iterEdges():
            src_k, tgt_k = g_node2scc[src][0], g_node2scc[tgt][0]
            if src_k != tgt_k:
                cond.addEdge(src_k, tgt_k, checkMultiEdge=True)
        return g, cond, g_node2scc, cond_node2scc


def parse(file: str) -> Automaton:
    """Wrapper around the base HOA parser."""
    __parser = HOAParser()
    input_string = Path(file).read_text()
    input_lines = Path(file).read_text().splitlines()
    input_string = "\n".join(
        line for line in input_lines
        if not line.startswith("controllable"))
    # Read in Strix controllable APs
    control_line = [x for x in input_lines if x.startswith("controllable")]
    if control_line:
        control_split = control_line[0].split(":")[1].split()
        control = set(int(x) for x in control_split)
    else:
        control = set()
    hoa_obj: HOA = __parser(input_string)
    return Automaton(hoa_obj, filename=file, ctrl=control)



import logging
from abc import ABC, abstractmethod
from pathlib import Path

import msgspec
import tomli

from ..drivers import CompositeDriver, Driver, UserDriver
from ..hoa import Automaton
from ..runners import (Bound, CompositeRunner, DetCompleteSingleRunner, Hook,
                       Quit, Runner, SingleRunner, UserChoice)
from .toml_v1 import TomlV1


class ConfigurationError(Exception):
    pass


class Configuration(ABC):
    @property
    @abstractmethod
    def driver(self):
        raise NotImplementedError

    @driver.setter
    @abstractmethod
    def driver(self, value: Driver):
        raise NotImplementedError

    @property
    @abstractmethod
    def runner(self) -> Runner:
        raise NotImplementedError

    @runner.setter
    @abstractmethod
    def runner(self, value: Runner):
        raise NotImplementedError

    @staticmethod
    def factory(fname: Path, a: list[Automaton], monitor: bool = False) -> "Configuration":  # noqa: E501
        """Load a configuration file from `fname`.

        Args:
            fname (Path): Path of the configuration file
            a (list[Automaton]): Automata to run
            monitor (bool, optional): True iff automata should be monitored \
                for acceptance. Defaults to False.

        Raises:
            ConfigurationError: Something wrong with the provided \
                configuration file

        Returns:
            Configuration: An object with configuration details.
        """
        if fname.suffix != ".toml":
            raise ConfigurationError(f"Unsupported config format {fname.suffix}")  # noqa: E501
        with open(fname, "rb") as conf_file:
            try:
                toml = tomli.load(conf_file)
                assert "hoax" in toml, "Missing mandatory section [hoax]"  # noqa: E501
                assert "version" in toml["hoax"], "Missing mandatory field [hoax].version"  # noqa: E501
                conf_version = toml["hoax"]["version"]
                if conf_version == 1:
                    conf = msgspec.convert(toml, type=TomlV1)
                    return TomlConfigV1(fname, conf, a, monitor)
                else:
                    raise ConfigurationError(f"Unsupported version {conf_version}")  # noqa: E501
            except (AssertionError, tomli.TOMLDecodeError, msgspec.ValidationError) as err:  # noqa: E501
                raise ConfigurationError(err) from None


class DefaultConfig(Configuration):
    """Class for the default HOAX configuration."""

    @property
    def runner(self) -> Runner:
        return self._runner

    @runner.setter
    def runner(self, value: Runner):
        self._runner = value

    @property
    def driver(self) -> Driver:
        return self._driver

    @driver.setter
    def driver(self, value: Driver):
        self._driver = value

    def __init__(self, a: list[Automaton], mon: bool = False) -> None:
        aps = list(set(ap for aut in a for ap in aut.get_aps()))
        runner, aut = (
            (SingleRunner, a[0]) if len(a) == 1 else (CompositeRunner, a))
        self.driver = UserDriver(list(aps))
        if len(a) == 1:
            self.runner: Runner = SingleRunner(aut=a[0], drv=self.driver, mon=mon)  # noqa: E501
        else:
            self.runner = CompositeRunner(automata=a, drv=self.driver, monitor=mon)  # noqa: E501
        self.runner.add_nondet_action(UserChoice())


class TomlConfigV1(Configuration):
    """Builds a configuration from a TOML file (version 1)"""

    @property
    def driver(self) -> Driver:
        return self._driver

    @driver.setter
    def driver(self, value: Driver):
        self._driver = value

    @property
    def runner(self) -> Runner:
        return self._runner

    @runner.setter
    def runner(self, value: Runner):
        self._runner = value

    def __init__(self, fname: Path, conf: TomlV1, a: list[Automaton],
                 monitor: bool = False) -> None:
        logging.getLogger().setLevel(logging.DEBUG)
        logging.getLogger().handlers.clear()
        for log_conf in conf.log:
            logging.getLogger().addHandler(log_conf.get_handler())
        self.fname = fname
        aps = list(ap for aut in a for ap in aut.get_aps())
        d = CompositeDriver()
        for drv_conf in conf.drivers():
            base_dir = Path(self.fname).parent
            drv = drv_conf.get_driver(aps, base_dir)
            d.append(drv)

        aps_left = [ap for ap in aps if ap not in set(d.aps)]
        if aps_left:
            default_driver = conf.hoax.get_default_driver()
            d.append(default_driver(aps_left))
        self.driver = d if len(d.drivers) > 1 else d.drivers[0]
        if len(a) > 1:
            self.runner: Runner = CompositeRunner(a, d, monitor)
        elif all(
            x in (a[0].hoa.header.properties or [])
            for x in ("complete", "deterministic")
        ):
            self.runner = DetCompleteSingleRunner(a[0], d, monitor)
        else:
            self.runner = SingleRunner(a[0], d, monitor)

        nondet_action = conf.runner.get_nondet()
        if nondet_action is not None:
            self.runner.add_nondet_action(nondet_action)
        if conf.runner.bound > 0:
            bound_cond = Bound(conf.runner.bound)
            self.runner.add_transition_hook(
                Hook(bound_cond, Quit(cause=bound_cond)))

    def get_driver(self):
        return self.driver



import importlib.metadata
import sys
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from pathlib import Path
from signal import SIG_DFL, SIGPIPE, signal
from typing import Annotated, Optional

import typer

from .config.config import Configuration, DefaultConfig
from .drivers import EndOfFiniteTrace
from .hoa import parse
from .runners import StopRunner

signal(SIGPIPE, SIG_DFL)

app = typer.Typer()


def print_version(version: bool):
    if version:
        print(f"hoax {importlib.metadata.version("hoax-hoa-executor")}")
        raise typer.Exit()


@app.command()
def hoax(
        files: list[Path],
        config: Annotated[
            Optional[Path],
            typer.Option(help="Path to a TOML config file.")] = None,
        monitor: Annotated[
            bool,
            typer.Option(help="Monitor the automaton's acceptance condition.")
        ] = False,
        quiet: Annotated[
            bool,
            typer.Option(help="Suppress output of individual transitions.")
        ] = False,
        version: Annotated[
            Optional[bool],
            typer.Option(
                "--version",
                help="Print version number and exit.",
                callback=print_version, is_eager=True)] = None,
):
    """Execute HOA automata"""

    with ThreadPoolExecutor() as exc:
        automata = list(exc.map(parse, files))

    conf = (
        Configuration.factory(config, automata, monitor)
        if config is not None
        else DefaultConfig(automata, monitor))

    run = conf.runner

    t = datetime.now()
    run.init()
    try:
        if quiet:
            while True:
                run.step()
        else:
            while True:
                tr = run.step()
                if tr:
                    print(*(f"{i}: {t}" for i, t in enumerate(tr)), sep="\n")
    except (StopRunner, KeyboardInterrupt, EndOfFiniteTrace) as e:
        print(f"Stopping due to {repr(e)}", file=sys.stderr)
    print(run.count, "steps,", datetime.now() - t, "seconds", file=sys.stderr)



from itertools import chain, combinations
import logging
from pathlib import Path
import sys
from typing import Annotated, Collection, Optional

from msgspec import Meta, Struct, field

from ..hoa import extract_aps
from .. import drivers, runners


def invalid(field: str, valid: Collection[str]):
    valid = ', '.join(valid)
    raise ValueError(f"{field} must be one of {valid}") from None


def resolve(path: str, base_path: Path) -> Path:
    resolved = Path(path)
    if not resolved.is_absolute():
        resolved = (base_path / resolved).resolve()
    return resolved


class TomlV1(Struct):
    """The HOAX TOML configuration format, v1.

    This class defines a `msgspec` structured format, which directly \
    encodes to/decodes from TOML.
    """

    class HoaxSection(Struct):
        DRIVERS = {
            "flip": drivers.RandomDriver,
            "user": drivers.UserDriver,
            "json": drivers.JSONDriver,
            "txt": drivers.SimpleTxtDriver}

        version: Annotated[int, Meta(ge=1, le=1)]
        name: Optional[str] = None
        default_driver: Optional[str] = field(name="default-driver", default="user")  # noqa: E501

        def get_default_driver(self):
            return self.DRIVERS[self.default_driver]

        def __post_init__(self):
            if self.default_driver not in self.DRIVERS:
                invalid("default-driver", self.DRIVERS)

    class LogSection(Struct):
        LOG_LEVELS = {
            "none": logging.FATAL,
            "error": logging.ERROR,
            "warning": logging.WARNING,
            "info": logging.INFO,
            "debug": logging.DEBUG}
        name: Optional[str] = None
        level: Optional[str] = field(default="info")

        def get_level(self) -> int:
            return self.LOG_LEVELS[self.level or "info"]

        def get_handler(self) -> logging.Handler:
            handler = (
                logging.StreamHandler(sys.stdout)
                if self.name is None
                else logging.FileHandler(self.name))
            # TODO add format customization
            handler.setFormatter(logging.Formatter(logging.BASIC_FORMAT))
            handler.setLevel(self.get_level())
            return handler

        def __post_init__(self):
            if self.level not in self.LOG_LEVELS:
                invalid("level", self.LOG_LEVELS)

    class DriverSection(Struct):
        class RandomDriver(Struct):
            aps: set[str | int]
            bias: Annotated[float, Meta(ge=0, le=1)] = -1

            def get_driver(self, aps, _) -> drivers.RandomDriver:
                result = drivers.RandomDriver(extract_aps(aps, self.aps))
                if self.bias >= 0:
                    result.cum_weights = (self.bias, 1)
                return result

        class UserDriver(Struct):
            aps: set[str | int]

            def get_driver(self, aps, _) -> drivers.UserDriver:
                return drivers.UserDriver(extract_aps(aps, self.aps))

        class JSONDriver(Struct):
            aps: set[str | int]
            filename: str

            def get_driver(self, aps, base_path: Path) -> drivers.JSONDriver:
                stream = open(resolve(self.filename, base_path))
                return drivers.JSONDriver(extract_aps(aps, self.aps), stream)

        class SimpleTxtDriver(Struct):
            aps: set[str | int]
            filename: str

            def get_driver(self, aps, base_path: Path) -> drivers.SimpleTxtDriver:    # noqa: E501
                stream = open(resolve(self.filename, base_path))
                return drivers.SimpleTxtDriver(extract_aps(aps, self.aps), stream)  # noqa: E501

        flip: list[RandomDriver] = field(default_factory=list)
        user: list[UserDriver] = field(default_factory=list)
        json: list[JSONDriver] = field(default_factory=list)
        txt: list[SimpleTxtDriver] = field(default_factory=list)

    class RunnerSection(Struct):
        NONDET_VALUES = {
            "first": None,
            "random": runners.RandomChoice(),
            "user": runners.UserChoice()}

        bound: Annotated[int, Meta(gt=0)] = 0
        nondet: Optional[str] = field(default="first")

        def get_nondet(self) -> runners.Action | None:
            return self.NONDET_VALUES[self.nondet or "first"]

        def __post_init__(self):
            if self.nondet not in self.NONDET_VALUES:
                invalid("nondet", self.NONDET_VALUES)

    hoax: HoaxSection = field(name="hoax")
    driver: DriverSection = field(default_factory=DriverSection)
    runner: RunnerSection = field(default_factory=RunnerSection)
    log: list[LogSection] = field(default_factory=list)

    def __post_init__(self):
        dups = set()
        for d1, d2 in combinations(self.drivers(), 2):
            dups.update(d1.aps.intersection(d2.aps))
        if dups:
            raise ValueError(f"APs have multiple drivers: {dups}")

    def drivers(self):
        yield from chain(self.driver.flip, self.driver.user, self.driver.json, self.driver.txt)  # noqa: E501
